import java.lang.*;
import java.io.*;
import java.util.*;
import java.applet.*;
import java.awt.*;
import java.net.*;
import java.text.*;
import java.time.*;
import java.math.*;
import javax.swing.*;

class Solution \{
    public static void main(String args[]) \{
        /**
        $prompt
        */